// Phil_G November 2024 - 10ch reeds emulation FHSS tx, has adjustable transit speed, reversing, trims, ATV, sweep, S/C escapement emulation, inactivity alarm, etc
// A0=aileron, A1=elevator, A2=rudder, A3=throttle, A4=unused, A5=trim, A7=transit speed. Always use this wiring regardless of PPM channel order.
// Each toggle has two resistors, one from pos to signal and one from neg to signal. Value & tolerance unimportant, must be same, say 10k 5%
// Analogue inputs are used purely to detect three toggle positions on one pin, they have no direct bearing on servo position.
// Single channel escapement emulation is sequential by grounding D8 and compound on D7 (one for right rudder, two for left, three for kick-up elevator)
// End-points set by link or momentary pushbutton on D6.  Remove electric props before setting end points!!!
// Optionally a Tiny-6 style trim button can be fitted to D5, in which case the trim toggle is not necessary.
// D2 and D3 select the four most common activetoggle configurations via a dipswitch. Channel order AETR or TAER is set in receiver.
// Please dont use the Arduino bootloader for programming, its startup delay means you can easily mess the configuration up.
// Follow the story on the single-channel forum
// Documentation with wiring diagram on http://www.singlechannel.co.uk  Archive page item P19
// If you use this software in whole or in part please acknowledge its source - Phil_G on most forums, philg@talk21.com  http://www.singlechannel.co.uk
// Single character Morse identifiers are used to confirm modes: D = reset to Defaults, R = Reversed, A = set ATV, V = ATV end, S = Sweep, K = OK.

/************* Radio config - make it unique! **********************************/
const byte pipeOut[] = "54321";   // last 5 digits of your phone number is sufficiently unique in our small group
const byte mychannels[] = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16};  // choose 16 FHSS hopping channels (0-82)

/*******************************************************************************/

// dont change any of these:
#define default_servo_value 1500  //set the default servo value
#define config_a 2
#define config_b 3
#define buzzer   4
#define t6trim   5
#define atvlink  6
#define sw_btn   7 // pushbutton for s/c compound
#define sw_seq   8 // pushbutton for s/c sequential
#define speedpot A7
#define cwspeed  75

#include <SPI.h>
#include <nRF24L01.h>
#include <RF24.h>

RF24 radio(10, 9);  // RF-NANO and new builds, ce=10, csn=9.   Phils original was RF24 radio(9, 10) dont use this for new builds
int rfchan = 0; // NRF channel index

// *** note NRF24L01 has a max 32 byte packet size ***
// Order not citical but receiver struct must match.
struct Packet { //    *** note NRF24L01 has a max 32 byte packet size ***
  unsigned int aileron  = 1500;
  unsigned int elevator = 1500;
  unsigned int throttle = 1100;
  unsigned int rudder   = 1500;
};

Packet data;
volatile int pktcounter = 0, sync = 0;
bool activetoggles[] = {1, 1, 1, 1, 0, 1}; // wired toggles, 0 for no, 1 for yes in the order:  0aileron, 1elevator, 2rudder, 3throttle, 4aux, 5trim. (Aux unused here)
volatile int ch[6];
int i, temp, dir = 1, val = 1, j, k, key = 0, increment;
int toggles[]  = {1, 1, 1, 1, 1, 1}; // ail, ele, rud, thr, aux, trim.  0=left,1=neutral,2=right. Aux unused here
int positions[] = {1400, 1400, 1400, 1000, 1000, 1400}; // neutral minus trim
int atvpos[] = {1800, 1800, 1800, 1800, 1800, 1800}; // positions at the time atv is set
int atv[] = {400, 400, 400, 400, 400, 400}; // preset travel from neutral, per channel  ail, ele, rud, thr, aux, (aux & aux2 unused)
int targets[]  = {1400, 1400, 1400, 1000, 1000, 1400}; // neutral minus trim
byte trims[]  = {10, 10, 10, 10, 10, 10};
byte reverse[] = {0, 0, 0, 0, 0, 0};
byte id[] = {0, 3, 1, 1, 0, 1, 0, 0, 0, 3, 3, 1, 0, 1, 1, 1, 1, 3, 0, 1, 3, 3, 1, 0, 1, 1, 1, 1, 0, 1, 3, 1, 1, 0};
static int rotor[] = {0, 36, 72, 108, 144, 180, 216, 252, 216, 180, 144, 108, 72, 36, 0, -36, -72, -108, -144, -180, -216, -252, -216, -180, -144, -108, -72, -20}; // last should strictly be -36
byte rotorptr = 0, flip = 0, flipped = 0, sequence = 1, trimlock = 0, atvflag = 0, anykey = 0;
int scrudder = 0, scelev = 0, mult;
unsigned int inact = 0; //inactivity frame counter. 30000 x 20ms frames is 10 minutes.
unsigned int transit, channelinc[] = {20, 20, 20, 20, 20, 20};

#include <EEPROM.h>
// eeprom addresses (bytes).  No attempt to be compact, reverse bit flags stored as bytes.
// 00
// initflg, if 0x5a if eeprom has been initialised, 255 if not

// 01      02      03      04      5       (6)
// ailtrim eletrim rudtrim thrtrim auxtrim (auxtrim & aux2trim-unused)

// 7        8       9       10      11      (12)
// ailrev   elerev  rudrev  thrrev  auxrev  (auxrev & aux2rev-unused)

// 13       14      15      16      17      (18)
// ailatv   eleatv  rudatv  thratv  auxatv  (auxatv & aux2atv-unused)

void setup() {
  pinMode(config_a, INPUT_PULLUP);
  pinMode(config_b, INPUT_PULLUP);
  pinMode(t6trim, INPUT_PULLUP);
  pinMode(sw_btn, INPUT_PULLUP);
  pinMode(sw_seq, INPUT_PULLUP);
  pinMode(atvlink, INPUT_PULLUP);
  pinMode(buzzer, OUTPUT);

  // indicate which toggles are wired up, 0 for no, 1 for yes in the order:  0aileron, 1elevator, 2rudder, 3throttle, 4aux, 5trim. Aux unused here
  activetoggles[4] = 0; //  {1,1,1,1,0,1} 10ch default. Aux unused here
  if (digitalRead(config_a) == 1 && digitalRead(config_b) == 0) { // {1,1,0,0,0,1} 6ch rudder elevator trim
    activetoggles[2] = 0;
    activetoggles[3] = 0;
  } 
  if (digitalRead(config_a) == 0 && digitalRead(config_b) == 1) { // {1,1,0,1,0,0} 6ch rudder elevator throttle, button trim
    activetoggles[2] = 0;
    activetoggles[5] = 0;
  } 

  read_toggles();
  if (EEPROM.read(0) != 0x5a  || toggles[5] == 2) {
    EEPROM.write(0, 0x5a); // initialise eeprom in location 0, ox5a is initialised key
    for (j = 1; j < 7; j++)  EEPROM.write(j, 10); // set all trims midway
    for (j = 7; j < 13; j++) EEPROM.write(j, 0); // clear all reverse settings
    for (j = 13; j < 19; j++) EEPROM.write(j, 200); // clear all atv settings (half as eeprom is byte)
    pip(120, 40); pip(40, 40); pip(40, 40); // D for reset eeprom back to defaults
  }

  // read trims, reversing & atv from eeprom, set any reversing commands
  for (j = 1; j < 6; j++) {
    trims[j - 1] = EEPROM.read(j); // read trims
    atv[j - 1] = 2 * (EEPROM.read(j + 12)); // read atv
    reverse[j - 1] = EEPROM.read(j + 6) & 1; // get reverse settings, 0 or 1
    if (toggles[j - 1] != 1) {
      reverse[j - 1] ^= B00000001;
      EEPROM.write(j + 6, reverse[j - 1]);
      pip(40, 40); pip(120, 40); pip(40, 40); // R for reversed
    }
  }

  if (anykey == 0 && digitalRead(t6trim) ==1) {
    pip(120, 40); pip(40, 40); pip(120, 40); // K for eeprom ok
  }

  //initialize default channel values
  for (j = 0; j < 6; j++) ch[j] = default_servo_value;
  // throttle defaults low
  if (reverse[3] != 1) {
    ch[2] = 1500 - atv[3]; // 1100
    positions[3] = 1400 - atv[3]; // 1000;
    targets[3] = 1400 - atv[3]; // 1000;
  } else {
    ch[2] = 1500 + atv[3]; // 1900;
    positions[3] = 1400 + atv[3]; // 1800
    targets[3] = 1400 + atv[3]; // 1800
  }

  while (toggles[5] == 2) {
    read_toggles(); // wait for release
    delay(100);
    if (++temp > 100) {
      temp = 0;
      cwid();
    }
  }
  // Tx config
  radio.begin(); delayMicroseconds(500);
  radio.openWritingPipe(pipeOut); delayMicroseconds(500);
  radio.setAutoAck(false); delayMicroseconds(500);
  radio.setDataRate(RF24_250KBPS); delayMicroseconds(500);
  // low power for testing, change to MAX later
  radio.setPALevel(RF24_PA_MAX); delayMicroseconds(500); // RF24_PA_MIN, RF24_PA_LOW, RF24_PA_HIGH and RF24_PA_MAX
  radio.setChannel(mychannels[0]); delayMicroseconds(500);
  radio.stopListening(); // transmit mode

  // set up 5ms packet send interrupt, CTC mode, top = 78, prescaler to /1024
  TCCR2A = 0;        // reset Timer2 control reg A
  TCCR2A |= (1 << WGM21);     // set WGM21 bit, CTC mode
  TCCR2B |= (1 << CS22) | (1 << CS21) | (1 << CS20); // 1024 prescaler
  OCR2A = 78;                 // top = 78
  TIMSK2 |= (1 << OCIE2A);    // enable timer2 compare interrupt
  sei();        // enable global interrupts
  if (toggles[5] == 0 || digitalRead(t6trim)==0) rangesweep();
}

void loop() {

  // read and quantify the toggles, 0=left, 1=neutral, 2=right.
  read_toggles();

  // elevator trim key does elevator if used alone, trims other channels if held simultaneously
  if (toggles[5] != 1 || digitalRead(t6trim) == 0) { // if elevator trim key or T6 button thrown
    delay(100); // allow some tolerance for keys held 'simultaneously'
    read_toggles();
    key = 0; // 'key' flags if any AERTX toggle is thrown
    for (j = 0; j < 5; j++) { // check for simultaneous keys
      if (toggles[j] != 1) {
        key = 1;  // trimlock prevents ele trim if rud etc released early
        trimlock = 1;
      }
      if (toggles[j] == 2 && trims[j] < 21) { // limit to 21 steps ie 0 to 20
        ++trims[j]; if (trims[j] == 10) pip(200, 100); else pip(80, 220);
      }
      else if (toggles[j] == 0 && trims[j] > 0) {
        --trims[j]; if (trims[j] == 10) pip(200, 100); else pip(80, 220);
      }
      if (j < 3 && key != 0) {
        targets[j] = 1400;  // neutralise channel whilst trimming
        positions[j] = 1400;
      }
      if (j > 2 && key != 0) {
        targets[j] = positions[j]; // hold progressive channels whilst trimming
      }
    }

    if (digitalRead(t6trim) == 1) {
      // no simultaneous keys so its elevator trim
      if (toggles[5] == 2 && key == 0 && trimlock == 0 && trims[1] < 21) {
        ++trims[1];  // limit to 21 steps ie 0 to 20
        if (trims[1] == 10) pip(200, 100);
        else pip(80, 220);
      }
      if (toggles[5] == 0 && key == 0 && trimlock == 0 && trims[1] > 0) {
        --trims[1];  // limit to 21 steps ie 0 to 20
        if (trims[1] == 10) pip(200, 100);
        else pip(80, 220);
      }
    }
    for (j = 1; j < 7; j++)  EEPROM.write(j, trims[j - 1]); // store trims
    //    while(toggles[5] !=1) read_toggles(); // prevents (in) or allows (out) trim repeat. I like it to repeat.
    read_toggles(); if (toggles[5] == 1) trimlock = 0; // release trim lock if elevator trim key released
  }

  //  if its not a trim command, set servo targets, order is 0=aileron, 1=elevator, 2=rudder, 3=throttle, 4=aux. Aux unused here.
  if (toggles[5] == 1) for (j = 0; j < 5; j++) {
      if (digitalRead(atvlink) == 1) targets[j] = toggles[j] * atv[j] + 1400 - atv[j];
      else targets[j] = toggles[j] * 450 + 950;
    }

  // if setting atv make all channels progressive
  if (digitalRead(atvlink) == 0) {
    if (toggles[0] == 1) {
      targets[0] = positions[0];
      atvpos[0] = positions[0];
    }
    if (toggles[1] == 1) {
      targets[1] = positions[1];
      atvpos[1] = positions[1];
    }
    if (toggles[2] == 1) {
      targets[2] = positions[2];
      atvpos[2] = positions[2];
    }
  }
  //  make throttle progressive rather than self-centring
  if (toggles[3] == 1) {
    targets[3] = positions[3];
    atvpos[3] = positions[3];
  }

  //  increment channels according to transit speed pot, cap at max throw
  for (j = 0; j < 3; j++) {
    if (targets[j] > positions[j]) {
      positions[j] += channelinc[j];
      if (positions[j] > targets[j]) positions[j] = targets[j];
    }
    if (targets[j] < positions[j]) {
      positions[j] -= channelinc[j];
      if (positions[j] < targets[j]) positions[j] = targets[j];
    }
  }
  // throttle move at half speed
  for (j = 3; j < 6; j++) {
    if (targets[j] > positions[j]) {
      positions[j] += channelinc[j] / 2;
      if (positions[j] > targets[j]) positions[j] = targets[j];
    }
    if (targets[j] < positions[j]) {
      positions[j] -= channelinc[j] / 2;
      if (positions[j] < targets[j]) positions[j] = targets[j];
    }
  }

  // single-channel escapement emulation
  if (digitalRead(sw_btn) == 0 || digitalRead(sw_seq) == 0) inact = 0;
  scrudder = 0; scelev = 0;

  // SC compound button
  if (rotorptr == 0 && digitalRead(sw_btn) == 0) ++rotorptr; // if esc is at rest and we press the button, start the rotation
  if (rotorptr != 0) {
    mult = atv[0] / 4;
    scrudder = (rotor[rotorptr] * mult) / 100;
    scelev = 0;
    if (digitalRead(sw_btn) == 0 && (rotorptr == 6 || rotorptr == 20)) --rotorptr; // do pawl stops @ 90 and 270 degrees
    if (digitalRead(sw_btn) == 0 && rotorptr == 27) {
      --rotorptr;  // do pawl stops @ 355 degrees kue
      scelev = atv[1] / 2;
    }
    ++rotorptr; if (rotorptr == 28) rotorptr = 0;
  }

  // SC sequential button
  if (digitalRead(sw_seq) == 1) flipped = 0;
  else {
    if (flipped == 0) {
      flip ^= 1;
      flipped = 1;
    }
    scrudder = flip == 0 ? -atv[0]*2/3 : atv[0]*2/3;
  }

  // set ATV when atvlink grounded
  if (digitalRead(atvlink) == 0) {
    if (atvflag == 0) {
      pip(40, 40); pip(120, 40); // A for set atv
      atvflag = 1;
    }
  }
  if (atvflag == 1 && digitalRead(atvlink) == 1) {
    pip(40, 40); pip(40, 40); pip(40, 40); pip(120, 40); // V for atv saved
    atvflag = 0;
    for (j = 0; j < 6; j++) {
      temp = 1400 - atvpos[j]; // get offset whichever direction
      temp = abs(temp); // no maths permitted within abs
      if (temp > 100) {
        atv[j] = temp;
        temp /= 2;
        EEPROM.write(j + 13, temp); // store atv settings (half as eeprom is byte)
      }
    }
  }

  // channel reordering.  Raw is 0=aileron, 1=elevator, 2=rudder, 3=throttle, 4=aux. Aux unused here.
  // ensure throw+trim+scthrow doesnt exceed standard values
  ch[0] = positions[0] + (10 * trims[0]) + scrudder; if (reverse[0] == 1) ch[0] = 3000 - ch[0]; ch[0] = constrain(ch[0], 850, 2150); // ch0=ail
  ch[1] = positions[1] + (10 * trims[1]) + scelev;   if (reverse[1] == 1) ch[1] = 3000 - ch[1]; ch[1] = constrain(ch[1], 850, 2150); // ch1=ele
  ch[2] = positions[3] + (10 * trims[3]);            if (reverse[3] == 1) ch[2] = 3000 - ch[2];                                      // ch2=thr
  ch[3] = positions[2] + (10 * trims[2]);            if (reverse[2] == 1) ch[3] = 3000 - ch[3];                                      // ch3=rud

  data.aileron     = ch[0];
  data.elevator    = ch[1];
  data.throttle    = ch[2];
  data.rudder      = ch[3];

  while (sync == 0); // wait for whatever remains of 20ms (four packets)
  sync = 0;

  // inactivity timer. 10 mins is 30,000 x 20ms frames.
  if (++inact > 30000 && bitRead(inact, 3) == 1 && bitRead(inact, 5) == 1) digitalWrite(buzzer, HIGH); else digitalWrite(buzzer, LOW);

} // end main()


// read toggle positions eg for aileron, ailtoggle 0=left, 1=neutral, 2=right
// crude method is faster than "ailtoggle = map(analogRead(A0),0,1023,0,2)" etc
void read_toggles() {
  anykey = 0;
  for (int toggleswitch = 0; toggleswitch < 6; toggleswitch++)  {
    if (analogRead(toggleswitch) > 682) toggles[toggleswitch] = 2;
    else if (analogRead(toggleswitch) < 341) toggles[toggleswitch] = 0;
    else toggles[toggleswitch] = 1;
    if (activetoggles[toggleswitch] == 0) toggles[toggleswitch] = 1; // make unconnected toggles neutral
    if (toggles[toggleswitch] != 1) {
      inact = 0;
      anykey = 1;
    }
  }
  // set servo transit time taking account of ATV
  transit = map(analogRead(speedpot), 0, 1023, 8, 50); // read transit speed pot
  for (k = 0; k < 6; k++) {
    if (digitalRead(atvlink) == 0) channelinc[k] = 2; // slow if we're setting atv
    else
      channelinc[k] = (transit * atv[k]) / 400;
  }
}

// Buzzer
void pip(unsigned long duration, unsigned long silence) {
  digitalWrite(buzzer, HIGH);
  delay(duration);
  digitalWrite(buzzer, LOW);
  delay(silence);
}

void cwid() {
  for (j = 0; j < sizeof(id) - 1; j++) {
    if (id[j] == 0) delay(cwspeed * 2);
    digitalWrite(buzzer, HIGH);
    ch[3] += 70; // option, rudder servo taps out morse
    delay(cwspeed * id[j]);
    digitalWrite(buzzer, LOW);
    ch[3] -= 70; // option, rudder servo taps out morse
    delay(cwspeed);
  }
}

void rangesweep() {
  pip(40, 40); pip(40, 40); pip(40, 300); // S for sweep
  dir = 1;
  // waggle ailerons, keep elec throttle shut
  data.aileron     = 1500;
  data.elevator    = 1500;
  data.rudder      = 1500;
  data.throttle    = 1500 - atv[3]; // 1100

  while (toggles[3] == 1) {  // until throttle keyed
    while (sync == 0);
    sync = 0;
    read_toggles();
    data.aileron += (channelinc[0] * dir);
    if (data.aileron > 1480 + atv[0]) {
      data.aileron = 1480 + atv[0];
      dir = -1;
      pip(15, 5);
    }
    if (data.aileron <= 1520 - atv[0]) {
      data.aileron = 1520 - atv[0];
      dir = 1 ;
      pip(15, 5);
    }
  }
}

ISR(TIMER2_COMPA_vect) {   // send a packet every 5ms
  radio.write(&data, sizeof(Packet));
  rfchan = (rfchan + 1) % 16;
  radio.setChannel(mychannels[rfchan]);
  pktcounter = (pktcounter + 1) % 4; // 20ms
  if (pktcounter == 3) sync = 1;
}

// If you use this software in whole or in part please acknowledge its source - Phil_G on most forums, philg@talk21.com  http://www.singlechannel.co.uk
// Read the story here: http://mode-zero.uk/viewtopic.php?f=42&t=971    and its predecessor here: http://mode-zero.uk/viewtopic.php?f=42&t=945
